package org.sc3d.apt.julia.v6;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/** A Canvas for displaying complex numbers. */
public class ArgandDiagram extends ImageCanvas {
  /** Constructs an ArgandDiagram, given values for its fields. The diagram is recorded on a BufferedImage of type 'BufferedImage.TYPE_BYTE_INDEXED' and with ColorModel 'ICM'. */
  public ArgandDiagram(Complex centre, double scale, int width, int height) {
    super(
      new BufferedImage(width, height, BufferedImage.TYPE_BYTE_INDEXED, ICM),
      width, height
    );
    this.centre = centre; this.scale = scale;
    this.pixels =
      ((DataBufferByte)this.image.getRaster().getDataBuffer()).getData();
  }
  
  /* New API. */
  
  /** The complex number at the centre of the diagram. */
  public final Complex centre;
  
  /** The size of a pixel of the diagram. */
  public final double scale;
  
  /** Equal to '((DataBufferByte)image.getRaster().getDataBuffer()).getData()'. */
  public final byte[] pixels;

  /** Returns the complex number corresponding to the centre of the specified pixel.
   * @param x the x-coordinate of the pixel (0 on left).
   * @param y the y-coordinate of the pixel (0 at top).
   * @return the complex number.
   */
  public Complex pixelToComplex(int x, int y) {
    return new Complex(
      this.centre.x + this.scale*0.5*(2*x+1-this.width),
      this.centre.y - this.scale*0.5*(2*y+1-this.height)
    );
  }
  
  /** Returns the x-coordinate of the pixel containing the specified complex number.
   * @param x the real part of the complex number. The imaginary part is not required.
   * @return the x-coordinate of the pixel. Values from '0' to 'width-1' are on the diagram.
   */
  public int complexToPixelX(double x) {
    return (int)Math.round(0.5*this.width + (x-this.centre.x)/scale);
  }
  
  /** Returns the y-coordinate of the pixel containing the specified complex number.
   * @param y the imaginary part of the complex number. The real part is not required.
   * @return the y-coordinate of the pixel. Values from '0' to 'height-1' are on the diagram.
   */
  public int complexToPixelY(double y) {
    return (int)Math.round(0.5*this.height - (y-this.centre.y)/scale);
  }
  
  /** Draws a point on the diagram. It does this by modifying 'image' and calling 'repaint()'. The on-screen version will be updated by the AWT soon afterwards.
   * @param z the point to plot.
   * @param colour the colour of the point. Only the bottom 8 bits are significant. They are interpreted as per 'ICM'.
   */
  public void drawPoint(Complex z, int colour) {
    final int x = this.complexToPixelX(z.x);
    final int y = this.complexToPixelY(z.y);
    if (x>=0 && x<this.width && y>=0 && y<this.height) {
      // Pixel is in range.
      this.pixels[x+this.width*y] = (byte)colour;
      this.repaint();
    }
  }
  
  /** Like 'drawPoint()', but draws a circle centred on the point.
   * @param z the point to plot.
   * @param radius the radius of the circle in pixels.
   * @param colour the colour.
   */
  public void drawFatPoint(Complex z, int radius, int colour) {
    final double r = radius * scale, r2 = r*r;
    final int xl = Math.max(this.complexToPixelX(z.x-r), 0);
    final int xr = Math.min(this.complexToPixelX(z.x+r), this.width-1);
    final int yt = Math.max(this.complexToPixelY(z.y+r), 0);
    final int yb = Math.min(this.complexToPixelY(z.y-r), this.height-1);
    if (xl<=xr && yt<=yb) {
      for (int y=yt; y<=yb; y++) for (int x=xl; x<=xr; x++) {
        if (this.pixelToComplex(x, y).sub(z).mod2() < r2)
          this.pixels[x+this.width*y] = (byte)colour;
      }
      this.repaint();
    }
  }
  
  /** Draws a vertical line on the diagram.
   * @param real the real part of a complex number on the line.
   * @param colour the colour of the line.
   */
  public void drawVerticalLine(double real, int colour) {
    final int x = this.complexToPixelX(real);
    if (x>=0 && x<this.width) {
      // Pixel is in range.
      for (int y=0; y<this.height; y++)
        this.pixels[x+this.width*y] = (byte)colour;
      this.repaint();
    }
  }
  
  /** Draws a horizontal line on the diagram.
   * @param imag the imaginary part of a complex number on the line.
   * @param colour the colour of the line.
   */
  public void drawHorizontalLine(double imag, int colour) {
    final int y = this.complexToPixelY(imag);
    if (y>=0 && y<this.height) {
      // Pixel is in range.
      for (int x=0; x<this.width; x++)
        this.pixels[x+this.width*y] = (byte)colour;
      this.repaint();
    }
  }
  
  /* Statics. */
  
  /** The ColorModel used for the diagram, which defines 256 easily distinguishable colours. '0' is black and '255' is white. */
  public static final IndexColorModel ICM;
  
  static {
    final byte[] c1 = new byte[256], c2 = new byte[256], c3 = new byte[256];
    for (int i=0; i<256; i++) {
      c1[i] = (byte)(((i&2)>0?146:0) | ((i&16)>0?73:0) | ((i&128)>0?36:0));
      c2[i] = (byte)(((i&1)>0?146:0) | ((i&8)>0?73:0) | ((i&64)>0?36:0));
      c3[i] = (byte)(((i&4)>0?170:0) | ((i&32)>0?85:0));
    }
    ICM = new IndexColorModel(8, 256, c1, c2, c3);
  }
  
  /* Override things in Component. */
  
  /** Draws 'image'. */
  public void paint(Graphics g) {
    g.drawImage(this.image, 0, 0, null);
  }
  
  /** Calls 'paint(g)'. */
  public void update(Graphics g) {
    this.paint(g);
  }
  
  /* Test code. */
  
  /** Returns a square ArgandDiagram of the specified size covering '-3.0' to '3.0' in each direction, and with vertical and horizontal lines at '-1.0', '0.0' and '1.0'. */
  public static ArgandDiagram makeStandardDiagram(int size) {
    final ArgandDiagram ans = new ArgandDiagram(
      Complex.ZERO, 6.0/size,
      size, size
    );
    ans.drawVerticalLine(-2.0, 2);
    ans.drawVerticalLine(-1.0, 2);
    ans.drawVerticalLine(0.0, 1);
    ans.drawVerticalLine(1.0, 2);
    ans.drawVerticalLine(2.0, 2);
    ans.drawHorizontalLine(-2.0, 2);
    ans.drawHorizontalLine(-1.0, 2);
    ans.drawHorizontalLine(0.0, 1);
    ans.drawHorizontalLine(1.0, 2);
    ans.drawHorizontalLine(2.0, 2);
    return ans;
  }
  
  /** Interprets arguments as points to plot on a diagram. The diagram stretches from '-3.0' to '3.0' in each direction, and is 512 pixels across. */
  public static void main(String[] args) {
    final ArgandDiagram me = makeStandardDiagram(512);
    for (int i=0; i<args.length; i++)
      me.drawFatPoint(Complex.valueOf(args[i]), 2, 255);
    me.display("Testing ArgandDiagram");
  }
}
