package org.sc3d.apt.julia.v6;

/** Represents a complex number. The real and imaginary parts are each stored as a double. Instances are immutable. */
public final class Complex {
  /** Constructs a Complex given its real and imaginary parts.
   * @param x the real part.
   * @param y the imaginary part.
   */
  public Complex(double x, double y) {
    this.x = x; this.y = y;
  }
  
  /** Constructs a Complex with an imaginary part of zero.
   * @param x the real part.
   */
  public Complex(double x) { this(x, 0.0); }

  /** Constructs a Complex given its real and imaginary parts as Strings.
   * @param x the real part.
   * @param y the imaginary part.
   * @throws NumberFormatException if 'x' or 'y' cannot be parsed as a double.
   */
  public Complex(String x, String y) throws NumberFormatException {
    this(Double.parseDouble(x), Double.parseDouble(y));
  }
  
  /** Attempts to construct a Complex by parsing a String in approximately the format returned by 'toString()'. More precisely, 's' must be of the form 'x + y i' in which 'x' and 'y' are Strings accepted by 'Double.valueOf()', "i" and "+" are literal characters, and white-space is ignored.
   * @param s the String to parse.
   * @throws NumberFormatException if 's' cannot be parsed.
   */
  public static Complex valueOf(String s) {
    s = s.trim();
    final int indexOfPlus = s.indexOf('+');
    if (indexOfPlus==-1 || s.charAt(s.length()-1)!='i') {
      throw new NumberFormatException(
        "\""+s+"\" does not contain both \"+\" and \"i\""
      );
    }
    return new Complex(
      s.substring(0, indexOfPlus),
      s.substring(indexOfPlus+1, s.length()-1)
    );
  }
  
  /* Static instances. */
  
  /** Equals to new Complex(0.0, 0.0). */
  public static final Complex ZERO = new Complex(0.0);

  /** Equals to new Complex(1.0, 0.0). */
  public static final Complex ONE  = new Complex(1.0);

  /** Equals to new Complex(0.0, 1.0). */
  public static final Complex I    = new Complex(0.0, 1.0);

  /** Equals to new Complex(0.0, -1.0). */
  public static final Complex NEGI = new Complex(0.0, -1.0);

  /** Equals to new Complex(0.0, 2*Math.PI). */
  public static final Complex TWOPII = new Complex(0.0, 2*Math.PI);
  
  /* Fields. */
  
  /** The real part of this Complex. */
  public final double x;
  
  /** The imaginary part of this Complex. */
  public final double y;
  
  /* One-operand operations. */
  
  /** Returns 'that!=null &amp;&amp; this.x==that.x &amp;&amp; this.y==that.y'. */
  public boolean equalsComplex(Complex that) {
    return that!=null && this.x==that.x && this.y==that.y;
  }
  
  /** Returns the squared modulus of 'this'. */
  public double mod2() {
    return this.x*this.x + this.y*this.y;
  }
  
  /** Returns this times 'I'. This method is the inverse of 'divI()'. */
  public Complex mulI() {
    return new Complex(-this.y, this.x);
  }
  
  /** Returns this times '-I'. This method is the inverse of 'mulI()'. */
  public Complex divI() {
    return new Complex(this.y, -this.x);
  }
  
  /** Returns the negative of 'this'. */
  public Complex neg() {
    return new Complex(-this.x, -this.y);
  }
  
  /** Returns the reciprocal of 'this'. Will cope even if 'this.mod2()' would overflow. */
  public Complex inv() {
    if (this.y>this.x ^ this.y>-this.x) {
      // |x| >= |y|, and x!=0.
      final double q = this.y / this.x;
      final double d = 1.0 / (this.x + q*this.y);
      return new Complex(d, -q*d);
    } else {
      // |x| <= |y|, and possibly y==0.
      if (this.y==0) return new Complex(1.0/this.x, 0.0);
      final double p = this.x / this.y;
      final double d = 1.0 / (p*this.x + this.y);
      return new Complex(p*d, -d);
    }
  }
  
  /** Returns the square of 'this'. Slightly more efficient than 'this.mul(this)'. This method is the inverse of 'sqrt()'. */
  public Complex sqr() {
    return new Complex((this.x+this.y)*(this.x-this.y), 2*this.x*this.y);
  }
  
  /** Returns the square-root of 'this'. This method is the inverse of 'sqr()' except that the real part of the answer will always be positive.
   * <p>This method calls 'mod2()' as part of the calculation, and so may not return the correct answer if that value overflows. I made this decision in order to reduce the number of calls to 'Math.sqrt()' from three to two.
   */
  public Complex sqrt() {
    if (this.x>0.0) {
      if (this.y==0.0) return new Complex(Math.sqrt(this.x), 0.0);
      final double ansX = Math.sqrt(0.5*(Math.sqrt(this.mod2())+this.x));
      return new Complex(ansX, 0.5*this.y/ansX);
    } else {
      if (this.y==0.0) return new Complex(0.0, Math.sqrt(-this.x));
      double ansY = Math.sqrt(0.5*(Math.sqrt(this.mod2())-this.x));
      if (this.y<0) ansY = -ansY;
      return new Complex(0.5*this.y/ansY, ansY);
    }
  }
  
  /** Returns the natural logarithm of 'this'. This method is the inverse of 'exp()', except that the imaginary part of the answer will be in the range from '-PI' to 'PI' (calculated using 'Math.atan2(this.y, this.x)').
   * <p>This method calls 'this.mod2()' and so will not give the correct answer if that value overflows.
   */
  public Complex log() {
    return new Complex(0.5*Math.log(this.mod2()), Math.atan2(this.y, this.x));
  }
  
  /** Returns 'E' raised to the power of 'this'. This method is the inverse of 'log()'. */
  public Complex exp() {
    final double r = Math.exp(this.x);
    return new Complex(r*Math.cos(this.y), r*Math.sin(this.y));
  }
  
  /** Returns the cosine of 'this', calulated as 'this.mulI().exp().add(this.divI().exp()).scale(0.5)'. */
  public Complex cos() {
    return this.mulI().exp().add(this.divI().exp()).scale(0.5);
  }
  
  /** Returns the sine of 'this', calculated as 'this.mulI().exp().sub(this.divI().exp()).divI().scale(0.5)'. */
  public Complex sin() {
    return this.mulI().exp().sub(this.divI().exp()).divI().scale(0.5);
  }
  
  /* Two-operand operations. */
  
  /** Returns 'this' plus 'that'. */
  public Complex add(Complex that) {
    return new Complex(this.x+that.x, this.y+that.y);
  }
  
  /** Returns 'this' minus 'that'. */
  public Complex sub(Complex that) {
    return new Complex(this.x-that.x, this.y-that.y);
  }
  
  /** Returns 'this' times 'that'. */
  public Complex scale(double that) {
    return new Complex(this.x*that, this.y*that);
  }
  
  /** Return 'this' times 'that'. */
  public Complex mul(Complex that) {
    return new Complex(
      this.x*that.x - this.y*that.y,
      this.x*that.y + this.y*that.x
    );
  }
  
  /** Returns 'this' divided by 'that', computed using 'this.mul(that.inv())'. */
  public Complex div(Complex that) {
    return this.mul(that.inv());
  }
  
  /** Returns 'this' raised to the power of 'that', computed using 'this.log().mul(that).exp()'. As a special case, 'ZERO' is returned if 'this' is zero but 'that' is non-zero. */
  public Complex pow(Complex that) {
    if (this.x==0 && this.y==0 && (that.x!=0 || that.y!=0)) return ZERO;
    return this.log().mul(that).exp();
  }
  
  /* Override things in Object. */
  
  /** Returns 'true' only if 'that' is an instance of Complex and 'this.equalsComplex(that)'. */
  public boolean equals(Object that) {
    try { return this.equalsComplex((Complex)that); }
    catch (ClassCastException e) { return false; }
  }
  
  /** Returns a hash code equal to 'h+846584375*i' where 'h' and 'i' are the hash codes of 'x' and 'y' as computed using 'java.lang.Double.hashCode()'. */
  public int hashCode() {
    return       new Double(this.x).hashCode() +
      846584375*(new Double(this.y).hashCode());
  }
  
  public String toString() {
    return this.x+" + "+this.y+"i";
  }
  
  /* Test code. */
  
  /** Does some calculations using complex numbers. */
  public static void main(String[] args) {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.julia.v6.Complex x+yi"
    );
    final Complex me = Complex.valueOf(args[0]);
    System.out.println("me = "+me);
    System.out.println("me.equalsComplex(null) = "+me.equalsComplex(null));
    System.out.println("me.equalsComplex(ONE) = "+me.equalsComplex(ONE));
    System.out.println("me.equalsComplex(me) = "+me.equalsComplex(me));
    System.out.println("me.mod2() = "+me.mod2());
    System.out.println("me.mulI() = "+me.mulI());
    System.out.println("me.divI() = "+me.divI());
    System.out.println("me.neg() = "+me.neg());
    System.out.println("me.inv() = "+me.inv());
    System.out.println("me.sqr() = "+me.sqr());
    System.out.println("me.sqrt() = "+me.sqrt());
    System.out.println("me.log() = "+me.log());
    System.out.println("me.exp() = "+me.exp());
    System.out.println("me.cos() = "+me.cos());
    System.out.println("me.sin() = "+me.sin());
    final Complex you = new Complex(2.0, 1.0);
    System.out.println("you = "+you);
    System.out.println("me.add(you) = "+me.add(you));
    System.out.println("me.sub(you) = "+me.sub(you));
    System.out.println("me.scale(2.0) = "+me.scale(2.0));
    System.out.println("me.mul(you) = "+me.mul(you));
    System.out.println("me.div(you) = "+me.div(you));
    System.out.println("me.pow(you) = "+me.pow(you));
  }
}
