package org.sc3d.apt.julia.v6;

import java.awt.event.*;
import java.awt.image.*;

/** An ImageCanvas that displays a Mandelbrot set. */
public class Mandelbrot extends ImageCanvas implements MouseListener, Runnable {
  /** Constructs a Mandelbrot, initially displaying the whole set.
   * @param size the size (width==height) of this Mandelbrot in pixels.
   */
  public Mandelbrot(int size) {
    super(
      new BufferedImage(size, size, BufferedImage.TYPE_INT_RGB),
      size, size
    );
    this.centre = new Complex(-0.5, 0);
    this.scale = 3.0/size;
    this.pixels =
      ((DataBufferInt)this.image.getRaster().getDataBuffer()).getData();
    this.recalculate();
    this.addMouseListener(this);
  }
  
  /* New API. */
  
  /** The point at the centre of this Mandelbrot. Do not change while calculating. */
  public Complex centre;
  
  /** The size of a pixel. Do not change while calculating. */
  public double scale;
  
  /** The colour frequency, in radians per factor of 'exp(1)'. Do not change while calculating. */
  public double freq = 6.0;
  
  /** Returns the Complex number corresponding to the centre of a pixel. */
  public Complex pixelToComplex(int x, int y) {
    return new Complex(
      this.centre.x + 0.5*this.scale*(2*x+1-this.width),
      this.centre.y - 0.5*this.scale*(2*y+1-this.height)
    );
  }
  
  /** Aborts the current calculation. This method blocks until the calculation Thread has terminated. Calls to 'abort()' and 'recalculate()' must strictly alternate. */
  public void abort() {
    if (this.thread==null) throw new IllegalStateException(
      "Must not call 'abort()' unless there is a current calculation"
    );
    this.shouldStop = true;
    try { this.thread.join(); }
    catch (InterruptedException e) {}
    this.thread = null;
  }
  
  /** Starts a new calculation. Do not call this until the previous calculation, if any, has terminated. */
  public void recalculate() {
    if (this.thread!=null) throw new IllegalStateException(
      "Must not call 'recalculate()' before previous calculation has finished"
    );
    // Start a new calculation Thread.
    this.shouldStop = false;
    this.thread = new Thread(this);
    this.thread.setPriority(Thread.MIN_PRIORITY);
    this.thread.start();
  }
  
  /* Implement things in Runnable. */
  
  /** Calculates this Mandelbrot calling 'repaint()' periodically and at the end. */
  public void run() {
    if (this.thread!=Thread.currentThread())
      throw new RuntimeException("Abandon ship!");
    // Print out the current position and scale.
    System.out.println("centre = "+this.centre+", pixel = "+this.scale);
    this.maxIt = (int)Math.max(Integer.MAX_VALUE, 1.0/this.scale);
    // Clear the Image.
    for (int i=0; i<this.pixels.length; i++) this.pixels[i] = BLANK;
    this.repaint();
    // Enqueue the borders of the Image.
    this.toDo = null;
    for (int x=0; !this.shouldStop && x<this.width; x++) {
      this.process(x, 0);
      this.process(x, this.height-1);
    }
    for (int y=0; !this.shouldStop && y<this.height; y++) {
      this.process(0, y);
      this.process(this.width-1, y);
    }
    // Pull pixels out of the queue and process their neighbours until done.
    long nextTime = System.currentTimeMillis();
    while (!this.shouldStop && this.toDo!=null) {
      if (System.currentTimeMillis()>=nextTime) {
        this.repaint();
        nextTime += 1000L;
      }
      final int x = this.toDo.x, y = this.toDo.y;
      this.toDo = this.toDo.next();
      for (int dx=-1; dx<=1; dx++) for (int dy=-1; dy<=1; dy++)
        if (dx!=0 || dy!=0) this.process(x+dx, y+dy);
    }
    // Fill in the black areas.
    for (int i=0; i<this.pixels.length; i++)
      if (this.pixels[i]==BLANK) this.pixels[i] = 0;
    this.repaint();
  }
  
  /* Implement things in MouseListener. */
  
  /** Re-centres on the selected point and zooms in by a factor of two. */
  public void mouseClicked(MouseEvent e) {
    this.abort();
    this.centre = this.pixelToComplex(e.getX(), e.getY());
    this.scale *= 0.3;
    this.recalculate();
  }
  
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mousePressed(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}
  
  /* Private. */
  
  /** The value used to represent uncalculated pixels. */
  private static final int BLANK = 0x7F7F7F7F;
  
  /** The array which backs 'this.image'. */
  private final int[] pixels;
  
  /** The Thread that is calculating this Mandelbrot, or 'null' if no calculation is in progress. */
  private Thread thread = null;
  
  /** Set to 'true' to request that 'this.thread' terminates soon. */
  private boolean shouldStop = false;
  
  /** The pixels that remain to be calculated. */
  private Queue toDo = null;
  
  /** The number of iterations at which to give up. This is calculated by 'run()' from 'scale'. */
  private int maxIt = 0;
  
  /** Calculates a pixel and adds it to 'toDo' unless it is black or off the screen or has already been calculated. Note that 'run()' calls this method, and that this method is not thread-safe, so you should not call this method while 'run()' is running. */
  private void process(int x, int y) {
    if (x<0 || x>=this.width || y<0 || y>=this.height) return;
    final int index = x+this.width*y;
    if (this.pixels[index]!=BLANK) return;
    // Calculate the pixel.
    final Orbit o = Orbit.make(
      Complex.ZERO,
      this.pixelToComplex(x, y),
      this.maxIt
    );
    final double pot = o==null ? Double.NEGATIVE_INFINITY : o.potential;
    if (pot==Double.NEGATIVE_INFINITY) {
      // Pixel is black.
      this.pixels[index] = 0;
      return;
    }
    // Pixel is coloured.
    final double a = this.freq * Math.log(10.0 - pot);
    this.pixels[index] = (int)(
      0x000001*Math.round(127.5*(1+Math.sin(a*0.9-2.1))) +
      0x000100*Math.round(127.5*(1+Math.sin(a*1.0+0.0))) +
      0x010000*Math.round(127.5*(1+Math.sin(a*1.1+2.1)))
    );
    // Add to 'toDo'.
    this.toDo = new Queue(pot, x, y).merge(this.toDo);
  }
  
  /** Represents a priority queue of pixels, sorted from low to high potential. 'null' represents the empty queue. */
  private static class Queue {
    /** Constructs a singleton queue. */
    public Queue(double pot, int x, int y) {
      this.pot = pot; this.x = x; this.y = y;
      this.p = null; this.q = null; this.depth = 1;
    }
    
    /** The potential of the first pixel. */
    public final double pot;
    
    /** A coordinate of the first pixel. */
    public final int x, y;
    
    /** Merges 'this' with 'that' destructively. */
    public Queue merge(Queue that) {
      if (that==null) return this;
      if (that.pot<this.pot) return that.merge(this);
      final int pd = this.p==null ? 0 : this.p.depth;
      final int qd = this.q==null ? 0 : this.q.depth;
      if (pd<qd) {
        this.p = that.merge(this.p);
        this.depth = 1 + (this.p.depth<qd ? this.p.depth : qd);
      } else {
        this.q = that.merge(this.q);
        this.depth = 1 + (pd<this.q.depth ? pd : this.q.depth);
      }
      return this;
    }
    
    /** Returns the Queue that remains after removing the first pixel. */
    public Queue next() {
      return this.p==null ? this.q : this.p.merge(this.q);
    }
    
    /** Two Queues which together include all the other pixels in this Queue. */
    private Queue p, q;
    
    /** One more than the minimum depth of 'p' and 'q'. 'null' is considered to have depth '0'. */
    private int depth;
  }
  
  /* Test code. */
  
  public static void main(String[] args) {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.julia.v6.Mandelbrot <size>"
    );
    final int size = Integer.parseInt(args[0]);
    new Mandelbrot(size).display("Testing Mandelbrot");
  }
}
