package org.sc3d.apt.julia.v6;

/** Represents the orbit of a point under the iteration 'z -> z2 + c', and provides useful statistics about it. */
public class Orbit {
  /** Constructs an Orbit, given values for its fields. */
  private Orbit(
    Complex z0, Complex c,
    int it, Complex zIt,
    double potential
  ) {
    this.z0 = z0; this.c = c;
    this.it = it; this.zIt = zIt;
    this.potential = potential;
  }
  
  /** Constructs an Orbit by dumb iteration. An Orbit is deemed to escape if it includes a point with a radius of more than '100'. An Orbit is deemed to be a prisoner if it revisits the same point more than once. Also, there are two special cases: Orbits starting at '0.0' with 'c' in the main cardioid or in the period-2 bud are detected and classed as prisoners immediately.
   * @param z0 the initial point on the Orbit.
   * @param c the constant in the mapping 'z -> z2 + c'.
   * @param maxIt the number of iterations to perform before giving up. A value of 'Integer.MAX_VALUE' is sensible but slow; smaller
   * @return a new Orbit, or 'null' if 'maxIt' iterations are not enough to determine whether or not the Orbit escapes.
   */
  public static Orbit make(final Complex z0, final Complex c, int maxIt) {
    if (z0.x==0.0 && z0.y==0.0) {
      // Mandelbrot point. Check special cases.
      if (
        Complex.ONE.add(c).mod2()<0.0625 ||
        Complex.ONE.sub(Complex.ONE.sub(c.scale(4)).sqrt()).mod2()<1.0
      ) {
        // Within period-2 bud or main cardioid.
        return new Orbit(z0, c, 0, z0, Double.NEGATIVE_INFINITY);
      }
    }
    double x = z0.x, y = z0.y;
    int it = 0;
    while (it<maxIt) {
      double loopX = x, loopY = y;
      int next = it + (it>>5) + 100; if (next<0) next = maxIt; // Overflow.
      while (it<next) {
        double r2 = x*x + y*y;
        if (r2>10000) {
          // Escapes.
          return new Orbit(
            z0, c,
            it, new Complex(x, y),
            Math.log(Math.log(r2)) / Math.log(2) - it
          );
        }
        double o;
        o = (x+y)*(x-y) + c.x; y = 2*x*y + c.y;
        x = (o+y)*(o-y) + c.x; y = 2*o*y + c.y;
        o = (x+y)*(x-y) + c.x; y = 2*x*y + c.y;
        x = (o+y)*(o-y) + c.x; y = 2*o*y + c.y;
        o = (x+y)*(x-y) + c.x; y = 2*x*y + c.y;
        x = (o+y)*(o-y) + c.x; y = 2*o*y + c.y;
        it += 6; if (it<0) return null; // Overflow
        if (x==loopX && y==loopY) {
          // Prisoner.
          return new Orbit(
            z0, c,
            it, new Complex(x, y),
            Double.NEGATIVE_INFINITY
          );
        }
      }
    }
    return null;
  }
  
  /* New API. */
  
  /** The initial point on the Orbit. */
  public final Complex z0;
  
  /** The constant in the mapping 'z -> z2 + c'. */
  public final Complex c;
  
  /** The number of iterations used to determine the fate of the orbit. */
  public final int it;
  
  /** The final point on the orbit. */
  public final Complex zIt;
  
  /** The log of the potential. The potential is defined to be the limit as 'it' tends to infinity of 'Math.log(zIt.mod2()) / Math.pow(2, it)'. The value of this field is therefore the limit as 'it' tends to infinity of 'Math.log(Math.log(zIt.mod2())) / Math.log(2) - it'. Orbits which do not escape to infinity will have a potential of '-Infinity'. */
  public final double potential;
  
  /* Test code. */
  
  /** Draws the Julia set for 'args[0]'. If the argument is omitted, picks a value at random on the boundary of the main cardioid or of the period-2 bud. */
  public static void main(String[] args) {
    Complex c;
    if (args.length==0) {
      double a = 2*Math.PI*Math.random();
      if (Math.random()<0.5) {
        c = new Complex(
          0.5*Math.cos(a) - 0.25*Math.cos(2*a),
          0.5*Math.sin(a) - 0.25*Math.sin(2*a)
        );
      } else {
        c = new Complex(
          0.25*Math.cos(a) - 1,
          0.25*Math.sin(a)
        );
      }
    } else if (args.length==1) {
      c = Complex.valueOf(args[0]);
    } else throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.julia.v6.Orbit [<real>+<imag>i]"
    );
    final ArgandDiagram ad = new ArgandDiagram(Complex.ZERO, 4.0/512, 512, 512);
    ad.display("Testing Orbit: c = "+c);
    for (int y=0; y<ad.height; y++) for (int x=0; x<ad.width; x++) {
      final Complex z0 = ad.pixelToComplex(x, y);
      final Orbit me = make(z0, c, 1000000);
      if (me!=null && me.potential>Double.NEGATIVE_INFINITY) {
        ad.drawPoint(z0, (int)Math.round(3*Math.log(2-me.potential)));
      }
    }
  }
}
