package org.sc3d.apt.julia.v6;

/** A diagram that shows the attractor of the mapping 'z -> +-sqrt(z - c)'. */
public class ReverseIteration extends ArgandDiagram {
  /** Constructs a ReverseIteration diagram. The diagram shows the area of the complex plane from '-2.0' to '2.0' in each direction.
   * @param c the complex number 'c' in the mapping 'z -> +-sqrt(z - c)'.
   * @param size the size of the diagram in pixels.
   */
  public ReverseIteration(Complex c, int size) {
    super(Complex.ZERO, 4.0/size, size, size);
    this.c = c;
    this.z = Complex.ONE.sub(c.scale(4.0)).sqrt().add(Complex.ONE).scale(0.5);
  }
  
  /** The constant passed to the constructor. */
  public final Complex c;
  
  /** A repelling fixed point. This point is guaranteed to be in the Julia set, on the boundary between prisoners and escapees. */
  public final Complex z;
  
  /** Draws the Julia set by reverse iteration staring at 'z1'. This method makes two recursive calls, passing 'z1.sub(c).sqrt()' and 'z1.neg().sub(c).sqrt()' respectively. For each call, 'enlarge' is multiplied by '4.0*z1.mod2()' and 'depth' is decremented by one. Recursion ends when 'enlarge>1.0' or 'depth<=0'.
   * @param z1 the first point (e.g. 'this.z').
   * @param enlarge the area scale factor desired (e.g. '0.1/(size*size)', or smaller to draw more points).
   * @param depth the recursion depth limit (e.g. 200, or larger to draw more points).
   * @return the number of points drawn.
   */
  public int draw(Complex z1, double enlarge, int depth) {
    Complex z2 = z1.neg();
    this.drawPoint(z1, 255); this.drawPoint(z2, 255);
    enlarge *= 4.0*z1.mod2();
    depth--;
    if (enlarge>1.0 || depth<=0) return 2;
    z1 = z1.sub(c); z2 = z2.sub(c);
    return 2 +
      this.draw(z1.sqrt(), enlarge, depth) +
      this.draw(z2.sqrt(), enlarge, depth);
  }
  
  /** Returns the point in the Julia set where the field line called 'alpha' lands. All points can be found by varying 'alpha'.
   * <p><em>This doesn't work!</em>
   */
  public Complex getPoint(long alpha) {
    System.out.println("getPoint(0x"+Long.toHexString(alpha)+")");
    Complex ans = this.z;
    for (long mask = alpha&-alpha; mask!=0L; mask<<=1) {
      ans = ans.sub(c).sqrt();
      if ((alpha&mask)!=0) ans = ans.neg();
      System.out.println("ans = "+ans);
    }
    return ans;
  }
  
  /** Draws the whole Julia set by passing a range of values to 'getPoint()'. The values are chosen using adaptive binary subdivision.
   * <p><em>This doesn't work!</em>
   */
  public int drawWholeSet() {
    this.drawPoint(this.z, 255);
    this.drawPoint(this.z.neg(), 255);
    return 2 + this.drawSet(Long.MIN_VALUE, this.z.neg(), 0L, this.z);
  }
  
  /* Private. */

  /** Draws the points returned by 'getPoint()' for a range of values of 'alpha', and also draws their mirror images.
   * <p><em>This doesn't work!</em>
   * @param a1 the start of the range (exclusive).
   * @param z1 equal to 'getPoint(a1)'.
   * @param a2 the end of the range (exclusive).
   * @param z2 equal to 'getPoint(a2)'.
   * @return the number of points drawn.
   */
  private int drawSet(
    final long a1, final Complex z1,
    final long a2, final Complex z2
  ) {
    System.out.println(
      "drawSet(\n"+
      "  0x"+Long.toHexString(a1)+", "+z1+",\n"+
      "  0x"+Long.toHexString(a2)+", "+z2+"\n)"
    );
    final double dist = Math.abs(z1.x-z2.x)+Math.abs(z1.y-z2.y);
    System.out.println("dist="+dist);
    if (dist<this.scale) return 0;
    final long aMid = a1 + ((a2-a1) >>> 1);
    if (aMid==a1) return 0; // Can't get any closer.
    final Complex zMid = this.getPoint(aMid);
    this.drawPoint(zMid, 255);
    this.drawPoint(zMid.neg(), 1);
    return 2 +
       this.drawSet(a1, z1, aMid, zMid) +
       this.drawSet(aMid, zMid, a2, z2);
  }
  
  /* Test code. */
  
  /** Draws a Julia set for 'args[0]', size 512. If no argument is given, uses a random point on the boundary of the main cardioid or of the period-2 bud. */
  public static void main(String[] args) {
    Complex c;
    if (args.length==0) {
      double a = 2.0*Math.PI*Math.random();
      if (Math.random()<0.5) {
        c = new Complex(
          0.5*Math.cos(a) - 0.25*Math.cos(2*a),
          0.5*Math.sin(a) - 0.25*Math.sin(2*a)
        );
      } else {
        c = new Complex(
          0.25*Math.cos(a) - 1.0,
          0.25*Math.sin(a)
        );
      }
    } else if (args.length==1) {
      c = Complex.valueOf(args[0]);
    } else throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.julia.v6.ReverseIteration <real>+<imag>i"
    );
    final int size = 512;
    final ReverseIteration me = new ReverseIteration(c, size);
    me.display("c = "+c);
    final int numPoints = me.draw(me.z, 0.01/(size*size), 200);
    System.out.println(numPoints+" points needed.");
  }
}
